/*
 Copyright 2016 Pagemesh Inc.

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */

module.exports = function(_) {
    const moment = require('moment');
    const OPERATORS = ['<', '>', '!', '*', '^', '~'];

    function buildSortCriteria(httpQuery) {

        var sort = _.pickBy(httpQuery, function(obj, key) {
            return key.indexOf('s_') === 0;
        });

        if (_.keys(sort).length > 0) {
            sort = _.map(sort, function(val, key) {
                return [key.substring(2), parseInt(val)];
            });
        } else {
            sort = [
                ['_id', -1]
            ];
        }

        return sort;
    }

    function replacePlaceholders(value, user) {
        if (_.isString(value)) {
            if (value.indexOf('$$') === 0) {
                if (value === '$$username') {
                    return _.get(user, "username");
                } else if (value === '$$uid') {
                    return _.get(user, "_id");
                }
            } else {
                return value;
            }
        } else {
            return value;
        }
    }

    function HttpQueryFactory(httpQuery, schema, options) {
        const excluded = ['apikey', 'limit', 'offset', 'sort', 'meta', 'radius'];

        options = options || {};

        var result = {
            op: 'find',
            opts: {}
        };

        result.sort = buildSortCriteria(httpQuery);
        result.query = _.pickBy(httpQuery, function(val, key) {
            return excluded.indexOf(key) === -1 && key.indexOf('s_') !== 0;
        });

        var types = {};
        if (schema) {
            types = schema.describe().children;
        }

        _.each(_.keys(result.query), function(key) {
            if (OPERATORS.indexOf(result.query[key][0]) !== -1) {
                let val;

                switch (result.query[key][0]) {
                    // Full-text search
                    case '*':
                        result.op = 'search';
                        result.query = result.query[key].substring(1);
                        break;
                        // Spatial Search
                    case '^':
                        var coordString = result.query[key].substring(1);
                        var lon = parseFloat(coordString.substring(0, coordString.indexOf(',')));
                        var lat = parseFloat(coordString.substring(coordString.indexOf(',') + 1));
                        result.op = 'search';
                        result.query = {
                            latitude: lat,
                            longitude: lon,
                            radius: httpQuery.radius
                        };
                        result.opts = {
                            type: 'spatial'
                        };
                        break;
                        // Partial string search
                    case '~':
                        result.op = 'search';
                        result.query[key] = {
                            $regex: new RegExp('^' + result.query[key].substring(1)),
                            $options: 'i'
                        };
                        result.opts = {
                            type: 'partial'
                        };
                        break;
                    case '>':
                        val = result.query[key].substring(1);
                        if (_.get(types[key], "type") === 'date') {
                            if (val[0] === '=') {
                                result.query[key] = {
                                    $gte: moment(val.substring(1)).toDate()
                                };
                            } else {
                                result.query[key] = {
                                    $gt: moment(val).toDate()
                                };
                            }
                        } else if (_.get(types[key], "type") === 'number') {
                            if (val[0] === '=') {
                                result.query[key] = {
                                    $gte: Number(val.substring(1))
                                };
                            } else {
                                result.query[key] = {
                                    $gt: Number(val)
                                };
                            }
                        } else {
                            if (val[0] === '=') {
                                result.query[key] = {
                                    $gte: replacePlaceholders(val.substring(1), options.user)
                                };
                            } else {
                                result.query[key] = {
                                    $gt: replacePlaceholders(val, options.user)
                                };
                            }
                        }
                        break;
                    case '<':
                        val = result.query[key].substring(1);
                        if (types[key].type === 'date') {
                            if (val[0] === '=') {
                                result.query[key] = {
                                    $lte: moment(val.substring(1)).toDate()
                                };
                            } else {
                                result.query[key] = {
                                    $lt: moment(val).toDate()
                                };
                            }
                        } else if (types[key].type === 'number') {
                            if (val[0] === '=') {
                                result.query[key] = {
                                    $lte: Number(val.substring(1))
                                };
                            } else {
                                result.query[key] = {
                                    $lt: Number(val)
                                };
                            }
                        } else {
                            if (val[0] === '=') {
                                result.query[key] = {
                                    $lte: replacePlaceholders(val.substring(1), options.user)
                                };
                            } else {
                                result.query[key] = {
                                    $lt: replacePlaceholders(val, options.user)
                                };
                            }
                        }
                        break;
                    case '!':
                        result.query[key] = {
                            $not: replacePlaceholders(result.query[key].substring(1))
                        };
                        break;
                }
            } else {
                result.query[key] = replacePlaceholders(result.query[key], options.user);
            }

        });

        return result;
    }

    return HttpQueryFactory;
};