/**
 * (C) 2016 Hub Group inc.
 */

module.exports = function(schema$_, Promise, _, Schema, log, injector, db$_) {

    function Service(schema) {
        if (_.isString(schema)) {
            schema = schema$_[schema];
        }
        this._schema = schema;
    }

    const _resolveRef = Promise.method(function(db, key, spec) {
        if (key) {
            if (_.isArray(key)) {
                let mongoList = injector.resolve('service$mongoList');
                return mongoList(db, spec.ref, {
                    _id: {
                        $in: key
                    }
                }, {
                    skipRef: true
                }).then(function(results) {
                    if (results && results.length > 0) {
                        if (spec.projection) {
                            return _.map(results, function(r) {
                                return _.pick(r, spec.projection);
                            });
                        } else {
                            return results;
                        }
                    } else {
                        return key;
                    }
                });
            } else {
                let filter = {};
                if (spec.key) {
                    filter[spec.key] = key;
                } else {
                    filter._id = key;
                }

                // Lazy resolution to avoid circular reference issue
                let mongoRead = injector.resolve('service$mongoRead');
                return mongoRead(db, spec.ref, filter).then(function(result) {
                    if (result) {
                        if (spec.projection) {
                            result = _.pick(result, spec.projection);
                        }
                        return result;
                    } else {
                        return key;
                    }
                });
            }
        }
    });

    const _updateRef = Promise.method(function(db, data, spec) {
        let mongoUpdate = injector.resolve('service$mongoUpdate');

        if (_.isArray(data)) {
            return Promise.map(data, function(obj) {
                let filter = {};
                if (!spec.key) {
                    spec.key = "_id";
                }
                filter[spec.key] = obj[spec.key];
                return mongoUpdate(db, spec.ref, filter, _.omit(obj, spec.key)).then(function(result) {
                    if (result) {
                        return result[spec.key];
                    }
                });
            });
        } else {
            let filter = {};
            if (!spec.key) {
                spec.key = "_id";
            }
            filter[spec.key] = data[spec.key];

            log.debug("Update ref %s", spec.ref, filter, _.omit(data, spec.key));
            return mongoUpdate(db, spec.ref, filter, _.omit(data, spec.key)).then(function(result) {
                if (result) {
                    return result[spec.key];
                }
            });
        }
    });

    Service.prototype.containerInfo = function(container) {
        if (this._schema) {
            let child = _.get(this._schema.describe().children, container);
            if (child && child.meta) {
                var info = _.zipObject(_.keys(child.meta), _.values(child.meta))[0];

                return _.merge(info, {
                    cardinality: child.type === 'array' ? 'many' : "one"
                });
            }
        }
    };

    Service.prototype.container = function(container) {
        if (this._schema) {
            var containerInfo = this.containerInfo(container);
            if (containerInfo && containerInfo.ref) {
                return new Service(schema$_[containerInfo.ref]);
            }
        }

        // Just return a schema service without any internal constraints
        return new Service();
    };

    Service.prototype.validate = function(objs, opts = {
        stripUnknown: true,
        allowUnknown: true,
        skipFunctions: true
    }) {
        let single = false;
        if (!_.isArray(objs)) {
            single = true;
            objs = [objs];
        }

        return Promise.map(objs, (o) => {
            if (this._schema) {
                let r = Schema.validate(o, this._schema, opts);
                if (r.error) {
                    throw r.error;
                } else {
                    return r.value;
                }
            } else {
                return o;
            }
        }).then(function(objs) {
            if (single) {
                return objs[0];
            } else {
                return objs;
            }
        });
    };

    Service.prototype.getSpatialIndexSpec = Promise.method(function() {
        var spatial = {};

        if (this._schema) {
            _.each(this._schema.describe().children, (value, key) => {
                if (value.meta) {

                    if (_.find(value.meta, function(m) {
                            return m.spatial;
                        })) {
                        var lat = 'latitude',
                            lng = 'longitude',
                            radius;

                        if (value.children) {
                            if (value.children.lat) {
                                lat = 'lat';
                            }

                            if (value.children.lon) {
                                lng = 'lon';
                            } else if (value.children.lng) {
                                lng = 'lng';
                            }

                            if (value.children.radius) {
                                radius = 'radius';
                            }
                        }

                        spatial[key] = '2dsphere';
                    }
                }

            });

            return spatial;
        } else {
            return spatial;
        }

    });

    Service.prototype.view = function(data, viewName, opts = {
        strict: false
    }) {
        if (this._schema) {
            _.each(this._schema.describe().children, function(field, key) {
                var pubSpec = _.find(field.meta, function(m) {
                    return m.hasOwnProperty(viewName);
                });

                if (opts.strict) {

                    if (!pubSpec || !pubSpec[viewName]) {
                        delete data[key];
                    }

                } else {
                    if (pubSpec && pubSpec[viewName] === false) {
                        delete data[key];
                    }
                }
            });
        }
        return data;
    };

    Service.prototype.resolveRefs = Promise.method(function(db, obj, opts = {}) {
        let result = obj;

        if (this._schema && !opts.skipRef) {
            let fields = this._schema.describe().children;
            return Promise.each(_.keys(fields), (key) => {

                let ref = _.find(fields[key].meta, function(m) {
                    return m.hasOwnProperty('ref');
                });

                if (ref) {
                    result[key] = _resolveRef(db, obj[key], ref);
                }

            }).then(function() {
                return Promise.props(result);
            });
        } else {
            return result;
        }
    });

    Service.prototype.createUniqueConstraint = Promise.method(function(data, skipConstraints) {
        var context = {};

        if (this._schema && !skipConstraints) {
            context.uniqueQuery = {
                $or: []
            };
            context.uniqueKeys = [];

            // Check all unique fields in database
            _.each(this._schema.describe().children, function(field, key) {
                if (_.find(field.meta, function(m) {
                        return m.hasOwnProperty('unique');
                    })) {
                    var q = {};
                    q[key] = data[key];
                    context.uniqueKeys.push(key);
                    context.uniqueQuery.$or.push(q);
                }
            });
            return context;
        } else {
            return {};
        }
    });

    Service.prototype.updateRefs = Promise.method(function(db, data, opts = {}) {
        if (this._schema) {

            let fields = this._schema.describe().children;
            return Promise.each(_.keys(fields), (key) => {

                let ref = _.find(fields[key].meta, function(m) {
                    return m.hasOwnProperty('ref');
                });

                // If the field is a ref and we have data, let's update this (if cascading) and replace with key
                if (data && ref && data[key]) {
                    if (_.isUndefined(opts.cascade) || opts.cascade.indexOf(ref.ref) !== -1) {
                        data[key] = _updateRef(db, data[key], ref);
                    } else {
                        let refObj = _.clone(data[key]);
                        if (_.isArray(refObj)) {
                            data[key] = _.map(refObj, function(o) {
                                return o[ref.key || '_id'];
                            });
                        } else {
                            data[key] = refObj[ref.key || '_id'];
                        }
                    }
                }

            }).then(function() {
                return Promise.props(data || {});
            });
        } else {
            return data;
        }
    });

    return function(key) {
        return new Service(key);
    };
};